@@ -23,12 +23,20 @@ import (
 	"bytes"
 	"encoding/json"
 	"errors"
+	"flag"
 	"fmt"
 	"go/format"
 	"io/ioutil"
 	"log"
+	"net/http"
 	"os"
 	"strings"
+	"unicode"
+)
+
+var (
+	uFlag = flag.Bool("u", false, "updates the debugProtocol.json file before generating the code")
+	oFlag = flag.String("o", "", "specifies the output file name. If unspecified, outputs to stdout")
 )
 
 // parseRef parses the value of a "$ref" key.
@@ -65,6 +73,9 @@ func parsePropertyType(propValue map[string]interface{}) string {
 		return parseRef(ref)
 	}
 
+	if _, ok := propValue["oneOf"]; ok {
+		return "interface{}"
+	}
 	propType, ok := propValue["type"]
 	if !ok {
 		log.Fatal("property with no type or ref:", propValue)
@@ -98,6 +109,8 @@ func parsePropertyType(propValue map[string]interface{}) string {
 			}
 			valueType := parsePropertyType(additionalProps.(map[string]interface{}))
 			return fmt.Sprintf("map[string]%v", valueType)
+		case "interface{}":
+			return "interface{}"
 		default:
 			log.Fatalf("unknown property type value %v in %v", propType, propValue)
 		}
@@ -279,7 +292,7 @@ func emitToplevelType(typeName string, descJson json.RawMessage) string {
 		} else if propName == "arguments" && (typeName == "LaunchRequest" || typeName == "AttachRequest") {
 			// Special case for LaunchRequest or AttachRequest arguments, which are implementation
 			// defined and don't have pre-set field names in the specification.
-			fmt.Fprintln(&b, "\tArguments map[string]interface{} `json:\"arguments\"`")
+			fmt.Fprintln(&b, "\tArguments json.RawMessage `json:\"arguments\"`")
 		} else {
 			// Go type of this property.
 			goType := parsePropertyType(propDesc)
@@ -293,6 +306,15 @@ func emitToplevelType(typeName string, descJson json.RawMessage) string {
 				// the specification) assume a value of 'true' for backward
 				// compatibility. See: https://github.com/google/go-dap/issues/39
 				jsonTag += "\"`"
+			} else if typeName == "InitializeRequestArguments" && (propName == "linesStartAt1" || propName == "columnsStartAt1") {
+				// These two special fields must not have the omitempty tag, despite being
+				// optional. If this attribute is missing the server will (according to
+				// the specification) assume a value of 'true'.
+				jsonTag += "\"`"
+			} else if typeName == "ErrorMessage" && propName == "showUser" {
+				// For launch/attach errors, vscode will treat omitted values the same way as true,
+				// so to supress visible reporting, we must report false explicitly.
+				jsonTag += "\"`"
 			} else {
 				jsonTag += ",omitempty\"`"
 			}
@@ -399,10 +421,59 @@ func emitMethodsForType(sb *strings.Builder, typeName string) {
 		fmt.Fprintf(sb, "func (e *%s) GetEvent() *Event {return &e.Event}\n", typeName)
 	}
 	if typeName == "LaunchRequest" || typeName == "AttachRequest" {
-		fmt.Fprintf(sb, "func (r *%s) GetArguments() map[string]interface{} { return r.Arguments }\n", typeName)
+		fmt.Fprintf(sb, "func (r *%s) GetArguments() json.RawMessage { return r.Arguments }\n", typeName)
 	}
 }
 
+func emitCtor(sb *strings.Builder, reqs, resps, events []string) {
+	fmt.Fprint(sb, `
+// Mapping of request commands and corresponding struct constructors that
+// can be passed to json.Unmarshal.
+var requestCtor = map[string]messageCtor{`)
+	for _, r := range reqs {
+		req := strings.TrimSuffix(firstToLower(r), "Request")
+		var msg string
+		if req == "initialize" {
+			msg = `
+	Arguments: InitializeRequestArguments{
+		// Set the default values specified here: https://microsoft.github.io/debug-adapter-protocol/specification#Requests_Initialize.
+		LinesStartAt1:   true,
+		ColumnsStartAt1: true,
+		PathFormat:      "path",
+},
+`
+		}
+		fmt.Fprintf(sb, "\n\t\"%s\":\tfunc() Message { return &%s{%s} },", req, r, msg)
+	}
+	fmt.Fprint(sb, "\n}")
+
+	fmt.Fprint(sb, `
+// Mapping of response commands and corresponding struct constructors that
+// can be passed to json.Unmarshal.
+var responseCtor = map[string]messageCtor{`)
+	for _, r := range resps {
+		resp := strings.TrimSuffix(firstToLower(r), "Response")
+
+		fmt.Fprintf(sb, "\n\t\"%s\":\tfunc() Message { return &%s{} },", resp, r)
+	}
+	fmt.Fprint(sb, "\n}")
+
+	fmt.Fprint(sb, `
+// Mapping of event ids and corresponding struct constructors that
+// can be passed to json.Unmarshal.
+var eventCtor = map[string]messageCtor{`)
+	for _, e := range events {
+		ev := strings.TrimSuffix(firstToLower(e), "Event")
+		fmt.Fprintf(sb, "\n\t\"%s\":\tfunc() Message { return &%s{} },", ev, e)
+	}
+	fmt.Fprint(sb, "\n}\n")
+}
+
+func firstToLower(s string) string {
+	r := []rune(s)
+	return string(unicode.ToLower(r[0])) + string(r[1:])
+}
+
 const preamble = `// Copyright 2020 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
@@ -417,12 +488,14 @@ const preamble = `// Copyright 2020 Google LLC
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-// DO NOT EDIT: This file is auto-generated.
+// Code generated by "cmd/gentypes/gentypes.go"; DO NOT EDIT.
 // DAP spec: https://microsoft.github.io/debug-adapter-protocol/specification
 // See cmd/gentypes/README.md for additional details.
 
 package dap
 
+import "encoding/json"
+
 // Message is an interface that all DAP message types implement with pointer
 // receivers. It's not part of the protocol but is used to enforce static
 // typing in Go code and provide some common accessors.
@@ -461,12 +534,12 @@ type EventMessage interface {
 type LaunchAttachRequest interface {
 	RequestMessage
 	// GetArguments provides access to the Arguments map.
-	GetArguments() map[string]interface{}
+	GetArguments() json.RawMessage
 }
 `
 
-// typesBlacklist is a blacklist of type names we don't want to emit.
-var typesBlacklist = map[string]bool{
+// typesExcludeList is an exclude list of type names we don't want to emit.
+var typesExcludeList = map[string]bool{
 	// LaunchRequest and AttachRequest arguments can be arbitrary maps.
 	// Therefore, this type is not used anywhere.
 	"LaunchRequestArguments": true,
@@ -474,9 +547,24 @@ var typesBlacklist = map[string]bool{
 }
 
 func main() {
+	flag.Parse()
+
 	log.SetFlags(log.LstdFlags | log.Lshortfile)
 
-	inputFilename := os.Args[1]
+	if flag.NArg() != 1 {
+		fmt.Fprintln(os.Stderr, "Path to the DAP specification json file is required.")
+		fmt.Fprintln(os.Stderr, "gentypes <path/to/debugProtocol.json>")
+		os.Exit(1)
+	}
+
+	inputFilename := flag.Arg(0)
+
+	if *uFlag {
+		if err := updateInput(inputFilename); err != nil {
+			log.Fatalf("Failed to update the input file: %v", err)
+		}
+	}
+
 	inputData, err := ioutil.ReadFile(inputFilename)
 	if err != nil {
 		log.Fatal(err)
@@ -499,19 +587,52 @@ func main() {
 		log.Fatal(err)
 	}
 
+	var requests, responses, events []string
 	for _, typeName := range typeNames {
-		if _, ok := typesBlacklist[typeName]; !ok {
+		if _, ok := typesExcludeList[typeName]; !ok {
 			b.WriteString(emitToplevelType(replaceGoTypename(typeName), typeMap[typeName]))
 			b.WriteString("\n")
 		}
 
 		emitMethodsForType(&b, replaceGoTypename(typeName))
+		// Add the typename to the appropriate list.
+		if strings.HasSuffix(typeName, "Request") && typeName != "Request" {
+			requests = append(requests, typeName)
+		}
+		if strings.HasSuffix(typeName, "Response") && typeName != "Response" && typeName != "ErrorResponse" {
+			responses = append(responses, typeName)
+		}
+		if strings.HasSuffix(typeName, "Event") && typeName != "Event" {
+			events = append(events, typeName)
+		}
 	}
 
+	// Emit the maps from id to response and event types.
+	emitCtor(&b, requests, responses, events)
+
 	wholeFile := []byte(b.String())
 	formatted, err := format.Source(wholeFile)
 	if err != nil {
 		log.Fatal(err)
 	}
-	fmt.Print(string(formatted))
+	if *oFlag == "" {
+		fmt.Print(string(formatted))
+	} else {
+		if err := ioutil.WriteFile(*oFlag, formatted, 0644); err != nil {
+			log.Fatalf("Failed to write the generated file: %v", err)
+		}
+	}
+}
+
+func updateInput(inputFilename string) error {
+	resp, err := http.Get("https://raw.githubusercontent.com/microsoft/vscode-debugadapter-node/main/debugProtocol.json")
+	if err != nil {
+		return err
+	}
+	defer resp.Body.Close()
+	data, err := ioutil.ReadAll(resp.Body)
+	if err != nil {
+		return err
+	}
+	return ioutil.WriteFile(inputFilename, data, 0644)
 }
